
class Maps:
    def __init__(self, map_types):
        self.types = self.get_map_data_type(map_types)

        # the dictionary
        self.__dict = {}

    # get the data types of the map
    def get_map_data_type(self, types):
        # check the type of the
        # types parameter and
        # return the tuple
        # as a sliced list
        if isinstance(types, tuple):
            return list(types[0:2])
        else:
            raise TypeError("Invalid type")
    
    # add an item to the 
    # map
    def add(self, add_item_key, add_item_value):
        # check whether the types
        # match to the specified
        # types and add it to
        # the dict
        if isinstance(add_item_key, self.types[0]) and isinstance(add_item_value, self.types[1]):
            if add_item_key not in self.__dict:
                self.__dict[add_item_key] = add_item_value
        else:
            raise TypeError("Types don't match")

    def get(self):
        return self.__dict

    def has(self, key_item):
        # check whether a specified
        # key is present in the dictionary
        # and return a boolean value
        return key_item in self.__dict

    # get the length of the map
    def length(self):
        return len(self.__dict)

    # delete the map item
    # based on the parameter
    # passed in
    def delete(self, delete_key):
        # check if the specified
        # key is present in the
        # map

        # if true delete it and return 
        # the map
        # else raise an Exception
        if delete_key in self.__dict:
            del self.__dict[delete_key]
            return self.get()
        else:
            raise Exception(f"Cannot find key {delete_key}")

    # filter the map
    # for a specific
    # value
    def filter(self, filter_value):
        # loop through each dict
        # keys and if the value 
        # match append it to the array
        # and return it
        find_arr = []

        for dict_key in self.__dict:
            if self.__dict[dict_key] == filter_value:
                find_arr.append(dict_key)
        return find_arr

    def clear(self):
        self.__dict = {}
        return self.get()

    # change a specific value
    def change(self, k, i):
        if k in self.__dict:
            if isinstance(i, self.types[1]):
                self.__dict[k] = i
        
        


# the number class
# the number data type
# can be integers, floats,
# doubles, longs stc

class Number():
    def __init__(self, number_value):
        # the main number value
        # integers or floats
        self.value = self.__get_value(number_value)

        # the number value
        # used instead of using
        # self.value['data']
        self.data = self.value['data']


    # get the value
    # of the number
    def __get_value(self, number):
        # check if the parameter
        # is an integer or a float
        # else try converting it
        # to a float and if an exception
        # occurs, throw out the Exception
        if isinstance(number, int) or isinstance(number, float):
            return self.__get_data_type(number)
        else:
            try:
                return self.__get_data_type(float(number))
            except Exception as number_convert_exception:
                raise number_convert_exception

    # convert the integer 
    # into string form
    def string(self):
        return str(self.value['data'])

    # get the data type
    # of the value passed in
    # integer, float, double, long
    def __get_data_type(self, data):
        return {
            "data" : data, # the number
            "type" : type(data), # the type , float or long
            "negative" : data < 0 # data is negative
        }

    # set or change the object value
    def set_(self, new_value):
        self.data = new_value
        self.value['data'] = new_value

        return new_value

    # fix the number
    # of decimal points
    # to the parameter
    def to_fixed(self, fix_point=0):  
        # make sure the type of
        # the parameter is int
        assert isinstance(fix_point, int), "Expected an integer"
        new_string = "" # create a new string
        points = str(self.data).split(".") # split on (.)
        new_string += f"{str(points[0])}." # add the first string   

        # if the length is greater
        # than 1(if value have decimal)
        # loop in range of the parameter
        # of the length of the decimal point
        # is less than the fix_point add some
        # zeros(0) at the end

        # return the string 
        if len(points) > 1:
            after_decimal_point = str(points[-1])
            for index in range(fix_point):
                if len(after_decimal_point) == (index + 1):
                    after_decimal_point += "0"
                
                new_string += after_decimal_point[index]
            return new_string
        else:
            return new_string
    
    # check whether the number is
    # an integer
    # via splitting the decimal
    # and verfying the
    # length
    @staticmethod
    def is_integer(number):
        return len(str(number).split(".")) == 1

    # get the value
    def get(self):
        return self.value["data"]

    def __repr__(self):
        return str(self.value["data"])


import string

# string data types
# strings are groups
# of characters

class String():
    def __init__(self, string_object):
        # the main string
        self.__string = str(string_object)

    # get the length of the 
    # string
    def length(self):
        return len(self.__string)

    # check whether a specified
    # substring is present
    # int the string
    def has(self, search_sub_string):
        return str(search_sub_string) in self.__string

    # remove a specific letter from the string
    def throw(self, throw_away_object):
        # check if the parameter is a string
        # if yes:pass
        # else , raise a TypeError
        # and check the length of the string

        # if the length == 1
        # create a string and append it to
        # if the current_character is not the target
        # and return the new variable

        # else raise and Exception
        if isinstance(throw_away_object, str):
            if len(throw_away_object) == 1:
                data = ""
                for index in range(len(self.__string)):
                    current_character = self.__string[index]
                    if current_character is not throw_away_object:
                        data += current_character
                return data

            else:
                raise Exception(f"Expected a character, but string of length {len(throw_away_object)} found")
        else:
            raise TypeError("Expected a string")

    # get the character 
    # at the specified index
    def character_at(self, index):
        if self.length() > index:
            return self.__string[index]
        else:
            raise Exception("Index out of range")

    # replace a specific character 
    # by annother word
    def replace(self, exists, new):
        # Create a new string
        # check whether both parameters
        # are strings

        # if yes loop through each element in 
        # the list and if the find the existsing 
        # string replace it with the new one

        # and return the new_string

        new_string = ""
        if isinstance(exists, str) and isinstance(new, str):
            for element in self.__string:
                if element == exists:
                    new_string += new
                else:
                    new_string += element
        else:
            raise TypeError("Arguments expected to be strings")
        return new_string

    # return the string in the form
    # of a list
    def each(self):
        return list(self.__string)
        
    # check for startswi5th or endswith
    def starts_with(self, element):
        if self.__string == "":
            return False
        else:
            return self.__string[0] == element

    def ends_with(self, element):
        if self.__string == "":
            return False
        else:
            return self.__string[-1] == element

    # remove all the spaces
    # from the string
    def remove_spaces(self):
        # create a new string
        # loop through each element
        # of the string
        # if the string is not a space
        # add it to the
        # new string
        new_string = ""
        for element in self.__string:
            if element is not " ":
                new_string += element
        return new_string

    # get the number of times 
    # a character appears
    # int the string
    
    def count(self, target):
        # loop through each
        # element in the
        # string and if the character
        # matched the target
        # increment the target_count

        target_count = 0

        if len(target) == 1:
            for el in self.__string:
                if el == target:
                    target_count += 1
            return target_count
        else:
            raise Exception(f"Expected a character, but string of length {len(target)} found")

    # get the index of the character
    def index(self, character):
        # if the length of the
        # string is 1 return
        # the index of the specified
        # element in the
        # string
        character = str(character)
        if len(character) == 1:
            return self.__string.index(character)
        else:
            raise Exception(f"Expected a character, but string of length {len(character)} found")

    # the repr function
    # is added inorder to 
    # return the private
    # string when calling the object

    # example => print(<string-object>)
    def __repr__(self):
        return self.__string


# the interface
# class, or the interface
# data type

# interface data types
# are inspired from
# the typescript language

# example
# f = Interface(types)
# f.create(data)

class InterfaceObject():
    def __init__(self, object_info):
        # the passed in values
        # to create a new object
        self.values = object_info

    # get the interface
    # property
    def get_item(self, property_):
        if property_ in self.values:
            return self.values[property_]
        else:
            raise KeyError(f"Cannot find {property_}")

    # set the interface
    # property

    # Note : cannot change types after assignment
    def set_item(self, property_, value):
        # check if property_ is in self.values
        # if yes, pass
        # else throw a KeyError
        if property_ in self.values:
            # check whether the current type
            # of the key and the new type is same
            # if yes , change the value
            # else , raise  a KeyError
            typeof = isinstance(value, type(
                self.values[property_]
            ))

            if typeof:
                self.values[property_] = value
            else:
                raise TypeError("Cannot change types after assignment")
        else:
            raise KeyError(f"Cannot find {property_}")

    def __repr__(self):
        return ""

class Interface(object):
    def __init__(self, data_params):
        # the main
        # key names and the
        # type of the value

        self.params = self.__get_params(data_params)


    # verifying the entered 
    # parameters and return
    # them
    def __get_params(self, data):
        # check whether the data
        # passed in is a dict
        # else throw an Error
        data_is_dict = isinstance(data, dict)

        # loop through each dict
        # value and check whether
        # each value is
        # a list
        # the list of types it can be
        # return the passed-in parameter
        if data_is_dict:
            for data_key in data:
                if not isinstance(data[data_key], list):
                    raise TypeError("Expected to be a list")
            return data
        else:
            raise TypeError("Expected a dict")

    # create a new interface
    # object
    def create(self, *args):
        # self.match = self.__match_param_types(data)
        # print(args)
        if len(args) == len(self.params):
            match = self.__match_param_types(args)
            return InterfaceObject(match)
        else:
            raise Exception(f"Expected {len(self.params)} arguments but got {len(args)}")

    # check whether
    # the new create interface
    # parameter type match
    # with self.params
    def __match_param_types(self, data):
        # check if the passed in
        # parameter is a 
        # tuple, else throw
        # a TYpeError
        if isinstance(data, tuple):
            # the new interface
            interface = {}
            # loop through each element 
            # and index in the dictionary
            # (self.params)
            for index, key in enumerate(self.params):
                # the current character
                # of the argument
                current_character = data[index]

                # the number of types
                # matches(a list of
                # True and False(booleans))
                types_matches = []
                for allowed_type in self.params[key]:
                    # if the allowed types
                    # is (?) or 'any', append true
                    # as it resembles (any) type

                    # else, if the type of
                    # the param is in
                    # the list of types
                    # allowed
                    if allowed_type == "?" or allowed_type == "any":
                        types_matches.append(True)
                    else:
                        types_matches.append(
                            allowed_type == type(data[index])
                        )
                # if the list is full of False
                # or if no types match raise
                # a type error
                # else, add it to the
                # interface dict
                if True not in types_matches:
                    raise TypeError("Types don't match")
                else:
                    interface[key] = data[index]
            # return the interface dict
            return interface
            
        else:
            raise TypeError("Expected a dict")



import json
from lol.database.database import Database
import dicttoxml

class DatabaseSerializer():
    def __init__(self, database_object):
        # the main database object
        # which contains the filename
        # and is converted to json data
        self.database = database_object

    # convert the database
    # values (in dict) to 
    # json form
    def jsonify(self):
        return json.dumps(self.database.get_dict_all_data_values_as_dict(), indent=6)

    # convert the database
    # values(int dict) to
    # xml form
    def xml(self):
        return dicttoxml.dicttoxml(self.database.get_dict_all_data_values_as_dict())





# the lol arrys
# the speciality of lol
# arrays is that
# they are typed and do not
# support duplicate elements
import math
import random

class Array():
    def __init__(self, typeof_array, length=None):
        # the type of the
        # elements in the array

        self.typeof = typeof_array

        # array members
        # this variable is private 
        # the user can only access
        # this via self.all()
        # function

        self.__array_members = []
        self.lengthof = length

    def all(self):
        # return all the members of
        # the array list
        return self.__array_members

    # add an element to the array
    def add(self, data):
        # check if the length specified
        # is none , if yes
        # pass else check if the size
        # limit is exceeded
        if self.lengthof is not None:
        
            if len(self.__array_members) == self.lengthof:
                raise Exception("Array size limit exceeded")

        # check the types
        if type(data) == self.typeof:
            # and append it to tha array
            # if it is not present in it
            if data not in self.__array_members:
                self.__array_members.append(data)
        else:
            raise TypeError("Type don't match")

    # remove an element from the
    # array based on
    # tyhe index specified
    def remove(self, index):
        if len(self.__array_members) > index:
            del self.__array_members[index]
            # return self.__array_members[index]
        return None

    # get the length of the array
    # an easy way for doing
    # len(<array-object>.all())
    def len(self):
        return len(self.__array_members)

    def element_at(self, index):
        # get the array element
        # at the specified index
        return self.__array_members[index]

    # get the median of the
    # array
    def median(self):
        """
        Get the median of the array
        => [1, 2, 3] => 
            i : floor(len(arr) / 2)
            -> arr[i]
        => [1, 2, 3, 4] =>
            i : floor(len(arr) / 2)
            -> ( arr[i] + arr[i - 1] ) / 2
        """
        if self.typeof == int or self.typeof == float:
            """
            if the length of the array
            is even get the sum of the 
            two numbers at the middle

            if the length of the array
            is odd get the number at the middle
            """
            if len(self.__array_members) % 2 == 0:
                arr_split_index = math.floor(self.len() / 2)
                return (self.__array_members[arr_split_index] + self.__array_members[arr_split_index - 1]) / 2
            else:
                return self.__array_members[
                    math.floor(self.len() / 2)
                ]

        else:
            raise Exception("elements should be numbers")

    # search insert position
    def search_inser_position(self, target_number):
        """
        Search insert position
        ----------------------
        => if element exists in arr return the
        index
        => else return the index if the element
        has to be appended to a sorted array
        """
        if target_number not in self.__array_members:
            self.add(target_number)

        return self.__array_members.index(target_number)

    # return the indexes
    def index(self, find_element):
        """
        Loop through all elements in
        the list and append to
        the index array
        if they find an element that match's
        the parameter
        """
        all_arr_indexes = []
        for index, element in enumerate(self.__array_members):
            if element == find_element:
                all_arr_indexes.append(index)

        return all_arr_indexes
        

    # return the first appearance 
    def first(self, find_element):
        arr = self.index(find_element)
        if len(arr) is not 0:
            return arr[0]
        
    # get the index of
    # the largest number in
    # the list
    def peak_index(self):
        return self.__array_members.index(
            max(self.all())
        )

    # get the count of the number
    # of times an element 
    # appears in the 
    # array
    def count(self, target_element):
        return len(self.index(target_element))

    # sort the array in ascending order
    def sort(self):
        self.__array_members.sort()
        return self.__array_members

    def find(self, d_t_f):
        indexes = []
        for x in range(len(self.__array_members)):
            if self.__array_members[x] == d_t_f:
                indexes.append(x)

        return indexes

    # clear the array
    def clear(self):
        self.__array_members = []
        return self.__array_members

    # reverse the array
    def reverse(self):
        self.__array_members.reverse()
        return self.__array_members

    # returns a multi-dimensional array
    # containing elements
    # at the alternate indexes
    def alternate_indexes(self):
        v_a_r = [[],[]]
        for x in range(len(self.__array_members)):
            if x % 2 == 0:
                v_a_r[0].append(x)
            else:
                v_a_r[-1].append(x)

        return v_a_r

    # return a random 
    # element from
    # the array
    def choice(self): 
        """
        Returns a randome element from the array
        """
        return self.__array_members[
            random.randint(
                0,
                self.len()
            )
            ]

    def front(self):
        if len(self.__array_members) is not 0:
            return self.__array_members[0]

    def back(self):
        if len(self.__array_members) is not 0:
            return self.__array_members[-1]

    def range(self, range):
        if range == "max":
            return max(self.__array_members)
        elif range == "min":
            return min(self.__array_members)
        else:
            raise Exception("Range should be max or min")

    def empty(self):
        return len(self.__array_members) == 0

    # extend the array size
    def extend(self, new_size):
        # if the length is not specified
        # or the length == NOne
        # check if the new_size is
        # greater than the old size
        # if yes asign the new value
        # to self.lengthog
        # else, throw an error
        if self.lengthof is not None:
            if new_size > self.lengthof:
                self.lengthof = new_size
            else:
                raise Exception("The new size should be greater than the current size")
        else:
            raise Exception("Cannot change the None to length")










import setuptools

# read the readme.md file and 
# add it as the
# long description of our package
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="loldb", # Replace with your own username
    version="0.0.3",
    author="P Pranav Baburaj",
    author_email="code-roller@googlegroups.com",
    description="A simple json database and other utilities",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pranavbaburaj/lol",
    packages=setuptools.find_packages(),
    install_requires=[
          'dicttoxml',
          "clint"
      ],
    entry_points = {
        "console_scripts": ['lol = cli.cli:main']
        },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


import os
import json
# the main database 

# uuid is used
# to create unique
# user id(s)
import uuid

# logging module
# is used log
# errors and warnings
import logging as logger

# the table class
# used to create
# many tables in the database
# each table is represented
# by  a json file
# in the name of the table

class Database():
    def __init__(self, table_name, fields):
        self.table = self.__get_table_name(table_name)
        self.fields = self.__get_field_name(fields)

        # create the database
        self.filename = self.__create_new_database_file(self.table)

        # index of the data
        # or id of each data fields
        self.idx = 0

        # the main dictionary in which the data
        # is going to be stored
        self.__data_dict = self.__get_file_dict()


        # print(self.__data_dict)

    
        self.__commit_additions()

        self.modifications = True

        # print(self.filename)

    # get the table name
    def __get_table_name(self, table_name):
        """
        if the table_name is a string
        return the string(with all spaces replaced by underscores)

        else raise a TypeError
        """
        if isinstance(table_name, str):
            return table_name.replace(" ", "_")
        else:
            raise TypeError("Table name expected to be a string")

    # get all valid fields
    def __get_field_name(self, fields):
        final_fields = []
        if isinstance(fields, list):
            # for dict_key in fields:
            #     final_fields.append(
            #         {
            #             "name" : dict_key,
            #             "typeof" : fields[dict_key]
            #         }
            #     )
            for index, el in enumerate(fields):
                if isinstance(el, str):
                    if el not in final_fields:
                        final_fields.append(el)
                else:
                    raise TypeError("Each field must be a string")
            return final_fields
        else:
            raise TypeError("Fields are expected to be dicts")

    def __create_new_database_file(self, database):
        # create the database
        # file for storing the information
        filename = os.path.join(os.getcwd(), f"{database}.lol")
        if not os.path.exists(filename):
            with open(filename, "w") as create_file_object:
                json.dump({}, create_file_object)
        
        
        return filename
    
    # commit the changes to the
    # file 
    # write the updated dictionary
    # to the file
    def __commit_additions(self):
        with open(self.filename, "w") as json_data_writer:
            json.dump(self.__data_dict, json_data_writer, indent=6)

    # add a new data to the 
    # the self.dictionary and
    # commit the changes
    # (write it to the file)
    def add(self, data_to_save):
        # check if the parameter
        # is a list object
        if isinstance(data_to_save, list):
            # check if the length of the field
            # match with the length of the 
            # parameter passed in
            # else, throw an Exception 
            check_data = self.__check_field_length(data_to_save)
            if check_data:
                new_dict = {}
                # for item in fields
                # change field item to key
                # and asign data[index]
                # to the dict key
                for index in range(len(self.fields)):
                    new_dict[self.fields[index]] = data_to_save[index]
                self.__data_dict[str(uuid.uuid4())] = new_dict

                # increment the index 
                # by 1
                self.idx += 1

                # commit the changes made to the dict
                self.__commit_additions()

                # create a logging data
                if self.modifications:
                    self.__create_message("Added data to the database")
        else:
            raise TypeError("Expected a list")

    # create info message
    # when doing actions
    def __create_message(self, data):
        print(f"INFO:{data}")

    # get the data inside of
    # the database file as json
    # or return the dict
    def __get_file_dict(self):
        with open(self.filename, "r") as json_reader:
            try:
                return json.load(json_reader)
            except Exception as exc:
                raise {}

    # get all the uuid's
    # in the data storing
    # dictioanry
    def ids(self):
        return [data_key for data_key in self.__data_dict]

    # clear all the fields
    # int the database
    # or empty the database
    def clear(self):
        # set the __data_dict to {}
        # and commit all the changes
        # to the file
        self.__data_dict = {}
        self.__commit_additions()

        if self.modifications:
            self.__create_message("Cleared the database")

    # delete a data field based
    # on the given key
    def delete(self, delete_object_key):
        if delete_object_key in self.__data_dict:
            del self.__data_dict[delete_object_key]
            self.__commit_additions()

            if self.modifications:
                self.__create_message("Deleted data from database")
        else:
            raise KeyError(f"Cannot find object with id {delete_object_key}")

    def set_track_modification(self, new_modify_value):
        if isinstance(new_modify_value, bool):
            self.modifications = new_modify_value
        else:
            raise TypeError("Expected a boolean value")

    # change a specific value
    # from the database
    def change(self, id, identifier, data):
        self.__data_dict[id][identifier] = data
        self.__commit_additions()

        if self.modifications:
            self.__create_message("Changed data in the database")

    # filter out objects
    # from the database dict based 
    # on the parameters
    # passed in
    def filter(self, matches):
        # make sure that the
        # parameter is a dict
        # else, throw a TypeError
        if isinstance(matches, dict):
            # check if the length
            # of the matches is equsl
            # to the length of the
            # fields
            data = self.__check_dict_length(matches)
            if data:
                fields = []
                for el in matches:
                    # append to the fields if it
                    # equal to the entered data
                    fields.append(self.__get_key_match(el, matches[el]))
                
                # return all duplicate items
                # from the fields
                return self.__remove_dup(fields)
        else:
            raise TypeError("Expected a dict")

    # return duplicat items
    def __remove_dup(self, fields):
        all_keys = []
        # take each element of
        # the first fields array and
        # make sure it is also present in
        # in the other array
        # if True:append to the final array
        for el in fields[0]:
            if self.__has_in_all(el, fields):
                all_keys.append(el)
        return all_keys

    # check if an element is present
    # in given array
    def __has_in_all(self, el ,fi):
        # for element in
        # fields
        # if el not present in one field
        # return false
        for k_fi in fi:
            if el in k_fi:
                pass
            else:
                return False
        return True
            

    # check whether it is same
    # as in the dictionary
    def __get_key_match(self, key, value):
        data = []
        for el in self.__data_dict:
            if self.__data_dict[el][key] == value:
                data.append(el)
        return data
        
        

    # get the object with a specific id
    def get(self, key):
        if key in self.__data_dict:
            return self.__data_dict[key]
        else:
            return {}

    # check whether the size
    # of the dict match with the matches
    def __check_dict_length(self, match):
        # loop through each element
        # in the dictionary and
        # make sure that the element
        # is present in the fields
        # else , raise a KeyError
        for el in match:
            if el not in self.fields:
                raise KeyError(f"Cannot find key {el}")
        return True
    
    # this is a secret function
    def get_dict_all_data_values_as_dict(self):
        return self.__data_dict


    # check whether the parameter
    # length is equal to the
    # self.fields length
    # else raise an error
    def __check_field_length(self, data):
        if len(data) == len(self.fields):
            return True
        else:
            raise Exception(f"Expected length {len(self.fields)} but got {len(data)}")